#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <sys/stat.h>
#include <process.h>
#include <direct.h>
#include "Atr.h"
#include "Utility.h"
#include "kboot.h"


int ConvertToHex( char *inFile, char *outFile)
{
FILE *in,*out;
int bytes,i;
const int bytesPerLine = 16;
unsigned char buffer[16];
unsigned long offset = 0;
in = fopen(inFile, "rb");
if ( !in )
    return 1;
out = fopen(outFile, "w");
if ( !out )
    {
    fclose(in);
    return 1;
    }
while ( (bytes = fread(buffer, sizeof(char), bytesPerLine, in)) > 0 )
    {
    fprintf( out, "%06lx:", offset);
    for(i=0;i<min(bytesPerLine,bytes);i++) 
        fprintf(out, " %02x", buffer[i] );
    fprintf(out, "\n");
    offset += bytesPerLine;
    if ( bytes < bytesPerLine )
        break;
    }
fclose(in);
fclose(out);
return 0;
}

int ConvertToAscii( char *inFile, char *outFile)
    {
    FILE *in,*out;
    int bytes,i;
    const int bufferSize = 2048;
    unsigned char buffer[2048];
    unsigned long offset = 0;
    in = fopen(inFile, "rb");
    if ( !in )
        return 1;
    out = fopen(outFile, "w");
    if ( !out )
        {
        fclose(in);
        return 1;
        }
    while ( (bytes = fread(buffer, sizeof(char), bufferSize, in)) > 0 )
        {
        for(i=0;i<min(bufferSize,bytes);i++) 
            fprintf(out, "%s", TranslateToAscii(buffer[i]) );
        offset += bufferSize;
        if ( bytes < bufferSize )
            break;
        }
    fclose(in);
    fclose(out);
    return 0;
    }

static char *statements[] = 
    {
    "REM", "DATA", "INPUT", "COLOR", "LIST", "ENTER", "LET",
    "IF", "FOR", "NEXT", "GOTO", "GO TO", "GOSUB", "TRAP",
    "BYE", "CONT", "COM", "CLOSE", "CLR", "DEG", "DIM",
    "END", "NEW", "OPEN", "LOAD", "SAVE", "STATUS", "NOTE",
    "POINT", "XIO", "ON", "POKE", "PRINT", "RAD", "READ",
    "RESTORE", "RETURN", "RUN", "STOP", "POP", "?",
    "GET", "PUT", "GRAPHICS", "PLOT", "POSITION", "DOS",
    "DRAWTO", "SETCOLOR", "LOCATE", "SOUND", "LPRINT",
    "CSAVE", "CLOAD", "", "ERROR- ",
    "DPOKE", "MOVE", "-MOVE", "*F", "REPEAT", "UNTIL", "WHILE",
    "WEND", "ELSE", "ENDIF", "BPUT", "BGET", "FILLTO", "DO",
    "LOOP", "EXIT", "DIR", "LOCK", "UNLOCK", "RENAME", "DELETE",
    "PAUSE", "TIME$=", "PROC", "EXEC", "ENDPROC", "FCOLOR", "*L",
    "-----------------------------------", "RENUM", "DEL", "DUMP",
    "TRACE", "TEXT", "BLOAD", "BRUN", "GO#", "#", "*B", "PAINT",
    "CLS", "DSOUND", "CIRCLE", "%PUT", "%GET", "??", "???", "????"

    };

/* first operator is 0x10 */
static char *operators[] =
    {
    "\"", "", ",", "$", ":", ";", "", " GOTO ", " GOSUB ", " TO ",
    " STEP ", " THEN ", "#", "<=", "<>", ">=", "<", ">", "=",
    "^", "*", "+", "-", "/", " NOT ", " OR ", " AND ", "(", ")",
    "=", "=", "<=", "<>", ">=", "<", ">", "=", "+", "-", "(",
    "", "", "(", "(", ","
    };

/* first function is 0x3D */
static char *functions[] =
    {
    "STR$", "CHR$", "USR", "ASC", "VAL", "LEN", "ADR", "ATN",
    "COS", "PEEK", "SIN", "RND", "FRE", "EXP", "LOG", "CLOG",
    "SQR", "SGN", "ABS", "INT", "PADDLE", "STICK", "PTRIG", "STRIG",
    "DPEEK", "&", "!", "INSTR", "INKEY$", "EXOR", "HEX$", "DEC",
    "DIV", "FRAC", "TIME$", "TIME", "MOD", "EXEC", "RND", "RAND",
    "TRUNC", "%0", "%1", "%2","%3", "GO#", "UINSTR", "ERR", "ERL",
    "??", "???", "????"
    };



int ConvertBasicToText(char *inFile, char *outFile, int typ)
    {
    FILE *in, *out;
    int token, slen;
    unsigned int bytes;
    unsigned char program[65534];
    unsigned char variables[32768], *varBufPtr = variables;
    unsigned char *varTable[128];
    unsigned char varBuf[256];
    unsigned char *varPtr, *bufPtr, *nextPtr, *stmtPtr, *stmtEndPtr;
    int lineNo, varCnt; //number of variables
    in = fopen(inFile, "rb");
    if ( !in )
        return 1;
    bytes = fread(program, sizeof(char), sizeof(program), in);
    if ( !bytes )
        return 1;
    fclose(in);
    if ( program[0] || program[1] )
        return 2;
    // write over the input file since it is a temp anyways 
    out = fopen(inFile, "w");
    if ( !out )
        return 1;
//    fprintf(out, "Variables:%c", 155);
    bufPtr = program+14;
    varCnt = 0;
    while ( *bufPtr )
        {
        varPtr = varBuf;
        while ( !((*bufPtr)&128) )
            *(varPtr++) = *(bufPtr++);
        *(varPtr++) = (*(bufPtr++))&127;
        *varPtr = 0;
        strcpy((char *)varBufPtr, (char *)varBuf);
        varTable[varCnt] = varBufPtr;
        varBufPtr+=strlen((char *)varBuf)+1;
        varCnt++;
 //       fprintf(out, "'%s'%c", varBuf, 155);
        
        }

//    fprintf(out, "%d variables%cProgram:%c%c", varCnt, 155, 155, 155);
    bufPtr = program + program[8] + program[9]*256 - 256 + 14;
    while ( (lineNo = (bufPtr[0] | (bufPtr[1]*256))) < 32768 )
        {
        nextPtr = bufPtr + (unsigned int)bufPtr[2];
        if ( !bufPtr[2] )
            {
            fprintf(out, "terminated because of zero offset%c", 155);
            break;
            }
        fprintf(out,"%u ", lineNo);
        stmtPtr = bufPtr + 3;
        while ( stmtPtr < nextPtr )
            {
            slen = *stmtPtr;
            stmtEndPtr = bufPtr+slen; //stmtPtr + slen;
            stmtPtr++; /* skip length byte */
            token = (*stmtPtr++);
            if ( token > 55 + 45 )
                token = 55 + 45;
            if ( strlen(statements[token])>0 )
                fprintf(out,"%s ", statements[token]);
            if ( token == 0 || token == 1 || token == 55 )
                {
                while ( *stmtPtr != 0x9B )
                    fprintf(out, "%c", *(stmtPtr++) );
                stmtPtr++;    
                break;
                }
            else
                {
                while( stmtPtr < stmtEndPtr)
                    {
                    char numBuf[64];
                    token = *(stmtPtr++);
                    if ( token == 0xE )
                        {
                        /* floating point constant */
                        Real6ToAscii( stmtPtr, numBuf );
                        fprintf(out, "%s", numBuf);
                        stmtPtr+=6;
                        }
                    else if ( token == 0xF )
                        {
                        /* string constant */
                        fprintf(out, "\"%*.*s\"", *stmtPtr, *stmtPtr,
                                stmtPtr+1);
                        stmtPtr += (*stmtPtr + 1);
                        }
                    else if ( token >= 0x10 && token <= 0x3C )
                        {
                        fprintf(out, "%s", operators[token-0x10]);
                        }
                    else if ( token >= 0x3D && token <= 0x54 + 25 )
                        {
                        fprintf(out, "%s", functions[token-0x3D]);
                        }
                    else if ( token >= 0x80 )
                        {
                        /* variable */
                        if (token-0x80>=varCnt)
                            fprintf(out,"??tok=%u??", token);
                        else
                            fprintf(out, "%s", varTable[token-0x80]);
                        }
                    else
                        {
                        fprintf(out,"??tok=%u??", token);
                        }




                    }
                }

            }
        fprintf(out, "%c", 155);
        bufPtr = nextPtr;
        if ( bufPtr - program > (int)bytes )
            {
            fprintf(out, "terminated because past end%c", 155);
            break;
            }
        }
    fprintf(out, "----------- end of file -----------%c", 155);
    fclose(out);

    
    ConvertToAscii(inFile,outFile);
    return(0);
    }

unsigned char *TranslateToAscii(char unsigned chr)
{
    static unsigned char str[2];
    unsigned char ch;
    str[0] = ch = chr&127; 
    if ( chr == 155 )
        return ((unsigned char *) "\n");
    if ( ch == '{' )
        return ((unsigned char *) "$(");
    if ( ch == '}' )
        return ((unsigned char *) "$)");
    if ( ch == '~' )
        return ((unsigned char *) "$-");
    if ( ch == 127 )
        return ((unsigned char *) "\t");
    if ( ch < ' ' )
        return ((unsigned char *) ".");
    if ( ch <= '_' )
        return ((unsigned char *) str );
    if ( ch < 'a' )
        return ((unsigned char *) ".");
    if ( ch <= '|' )
        return ((unsigned char *) str);
    return ((unsigned char *)".");
}


char * Real6ToAscii( unsigned char *real6, char *str)
{
    char tmpBuf[64];
    char *outPtr = str;
    char *tmpPtr = tmpBuf;
    int tmp;
    int i;
    int exponent, exponentIncrement = 0;

    *tmpPtr = *outPtr = 0;
    if ( real6[0] )
        {
        if ( (real6[0]&0x7F) >= 0x3F && (real6[0]&0x7F) < 0x45 )
            {
            for(i=1;i<6;i++)
                {
                *(tmpPtr++) = (char)((real6[i]>>4) + '0');
                *(tmpPtr++) = (char)((real6[i]&0xF) + '0');
                }
            tmpPtr = tmpBuf;
            tmp = 2*((real6[0]&0x7F) - 0x3F);
            *(outPtr++) = '0';
            for(i=0; i< tmp; i++)
                *(outPtr++) = *(tmpPtr++);
            *(outPtr++) = '.';
            for(i=tmp; i<10;i++)
                *(outPtr++) = *(tmpPtr++);
            *outPtr = 0;
            // strip trailing 0's and '.' if necessary
            outPtr --;
            while (*outPtr == '0' || *outPtr == '.' )
                {
                if ( *outPtr == '.' )
                    {
                    *outPtr = 0;
                    break;
                    }
                else
                    {
                    *(outPtr--) = 0;
                    }
                }
            // remove leading 0's if there is one
            while ( *str == '0' && *(str+1) != '.' )
                strcpy(str,str+1);
            }
        else
            {
            // display number in Fp format
            for(i=1;i<6;i++)
                {
                *(tmpPtr++) = (char)((real6[i]>>4) + '0');
                *(tmpPtr++) = (char)((real6[i]&0xF) + '0');
                }
            tmpPtr = tmpBuf;
            tmp = 2*((real6[0]&0x7F) - 0x3F);
            *(outPtr++) = *(tmpPtr++);
            if ( *(tmpPtr-1) == '0' )
                {
                *(outPtr-1) = *(tmpPtr++);
                i = 2;
                }
            else
                {
                exponentIncrement = 1;
                i=1;
                }
            *(outPtr++) = '.';
            for( ; i<10;i++)
                *(outPtr++) = *(tmpPtr++);
            *outPtr = 0;
            // strip trailing 0's and '.' if necessary
            outPtr --;
            while (*outPtr == '0' || *outPtr == '.' )
                {
                if ( *outPtr == '.' )
                    {
                    *(outPtr--) = 0;
                    break;
                    }
                else
                    {
                    *(outPtr--) = 0;
                    }
                }
            outPtr++;
            // now handle the exponent
            exponent = (real6[0]&0x7F)-64;
            if ( exponent < 0 )
                {
                *(outPtr++) = '-';
        //        exponent = -exponent;
                }
            else
                {
                *(outPtr++) = '+';
                }
            *(outPtr++) = 'E';
            exponent *=2;
            exponent += exponentIncrement;
            sprintf(outPtr, "%u", abs(exponent) );
            }

        }
    else
        {
        *(outPtr++) = '0';
        *outPtr = 0;
        }
    return str;

                


}

//-----------------------------------------------------------//
//   This function converts an XFD file to ATR format        //
//-----------------------------------------------------------//

int XFDtoATR( const char *xfdName, const char *atrName)
{
    FILE *xfd;
    _finddata_t fileInfo;
    long find;
    CAtrFile atr;
    unsigned char buffer[256];
    unsigned short sectors, secSize;
    unsigned short sec;

    find = _findfirst( (char *)xfdName, &fileInfo );
    if ( find == -1 )
        return 1;
    if ( fileInfo.size == 720*128 )
        {
        sectors = 720;
        secSize = 128;
        }
    else if ( fileInfo.size == 1040*128 )
        {
        sectors = 1040;
        secSize = 128;
        }
    else if ( fileInfo.size == 720*256 )
        {
        sectors = 720;
        secSize = 256;
        }
    else
        {
        return 3;
        }


    xfd = fopen(xfdName, "rb");
    if ( !xfd )
        return 1;
    if ( atr.Create( (char *) atrName, sectors, secSize) )
        {
        fclose(xfd);
        return 2;
        }
    
    for(sec = 1; sec <= sectors; sec++)
        {
        int bytes;
        bytes = fread(buffer, 1, secSize, xfd);
        if ( bytes != secSize )
            {
            atr.Close();
            fclose(xfd);
            return 4;
            }
        if ( atr.WriteSector(sec, buffer) )
            {
            atr.Close();
            fclose(xfd);
            return 5;
            }
        }
    fclose(xfd);
    atr.Close();
    return 0;
}

//-----------------------------------------------------------//
//   This function converts an ATR file to XFD format        //
//-----------------------------------------------------------//

int ATRtoXFD( const char *atrName, const char *xfdName)
{
    FILE *xfd;
    CAtrFile atr;
    unsigned char buffer[256];
    unsigned short sectors, secSize;
    unsigned short sec;
    unsigned char prot;

    if ( atr.Open((char *)atrName) )
        return 1;
    if ( atr.GetInfo( &secSize, &sectors, &prot) )
        {
        atr.Close();
        return 2;
        }
    if ( !((sectors == 720 && secSize == 256) ||
           (sectors == 720 && secSize == 128) ||
           (sectors == 1040 && secSize == 128)) )
        return 3;

    xfd = fopen(xfdName, "wb");
    if ( !xfd )
        {
        atr.Close();
        return 4;
        }
    
    for(sec = 1; sec <= sectors; sec++)
        {
        int bytes;
        if ( atr.ReadSector(sec, buffer) )
            {
            atr.Close();
            fclose(xfd);
            return 5;
            }
        bytes = fwrite(buffer, 1, secSize, xfd);
        if ( bytes != secSize )
            {
            atr.Close();
            fclose(xfd);
            return 6;
            }
        }
    fclose(xfd);
    atr.Close();
    return 0;
}

//-----------------------------------------------------------//
//   This function converts an ATR file to XFD format        //
//-----------------------------------------------------------//

int ATRtoNonStdXFD( const char *atrName, const char *xfdName)
{
    FILE *xfd;
    CAtrFile atr;
    unsigned char buffer[256];
    unsigned short sectors, secSize;
    unsigned short sec;
    unsigned char prot;

    if ( atr.Open((char *)atrName) )
        return 1;
    if ( atr.GetInfo( &secSize, &sectors, &prot) )
        {
        atr.Close();
        return 2;
        }
    if ( !((sectors == 720 && secSize == 256) ||
           (sectors == 720 && secSize == 128) ||
           (sectors == 1040 && secSize == 128)) )
        return 3;

    xfd = fopen(xfdName, "wb");
    if ( !xfd )
        {
        atr.Close();
        return 4;
        }
    
    for(sec = 1; sec <= sectors; sec++)
        {
        int bytes;
        if ( atr.ReadSector(sec, buffer) )
            {
            atr.Close();
            fclose(xfd);
            return 5;
            }
        bytes = fwrite(buffer, 1, (sec<4)?128:secSize, xfd);
        if ( bytes != ((sec<4)?128:secSize) )
            {
            atr.Close();
            fclose(xfd);
            return 6;
            }
        }
    fclose(xfd);
    atr.Close();
    return 0;
}

//-----------------------------------------------------------//
//   This function converts an XFD file to ATR format        //
//-----------------------------------------------------------//

int NonStdXFDtoATR( const char *xfdName, const char *atrName)
{
    FILE *xfd;
    _finddata_t fileInfo;
    long find;
    CAtrFile atr;
    unsigned char buffer[256];
    unsigned short sectors, secSize;
    unsigned short sec;

    find = _findfirst( (char *)xfdName, &fileInfo );
    if ( find == -1 )
        return 1;
    if ( fileInfo.size == 720*128 )
        {
        sectors = 720;
        secSize = 128;
        }
    else if ( fileInfo.size == 1040*128 )
        {
        sectors = 1040;
        secSize = 128;
        }
    else if ( fileInfo.size == 720*256-384 )
        {
        sectors = 720;
        secSize = 256;
        }
    else
        {
        return 3;
        }


    xfd = fopen(xfdName, "rb");
    if ( !xfd )
        return 1;
    if ( atr.Create( (char *) atrName, sectors, secSize) )
        {
        fclose(xfd);
        return 2;
        }
    
    for(sec = 1; sec <= sectors; sec++)
        {
        int bytes;
        if ( sec<4 )
            memset(buffer, 0 , 256);
        bytes = fread(buffer, 1, (sec<4)?128:secSize, xfd);
        if ( bytes != ((sec<4)?128:secSize) )
            {
            atr.Close();
            fclose(xfd);
            return 4;
            }
        if ( atr.WriteSector(sec, buffer) )
            {
            atr.Close();
            fclose(xfd);
            return 5;
            }
        }
    fclose(xfd);
    atr.Close();
    return 0;
}


/* converts a bad type double density ATR (Ape/XFormer) to a good
   ATR (Sio2PC, XLIT/ others */

int BadAtrtoGoodAtr( char *inFile , char *outFile)
    {
    int bytes,i;
    unsigned char buffer[256];
    unsigned long imageSize;
    FILE *in, *out;
    in = fopen(inFile, "rb");
    if ( !in )
        return 1;
    out = fopen(outFile, "wb");
    if ( !out )
        {
        fclose(in);
        return 2;
        }
    bytes = fread(buffer, sizeof(char), 16, in);
    if ( bytes != 16 )
            {
            fclose(in);
            fclose(out);
            return 3;
            }
    // correct the #of paragraphs in the header
    imageSize = 16L * (buffer[2] | buffer[3] * 256L | buffer[6] * 65536L);
    imageSize += 384;
    imageSize /= 16;
    buffer[2] = (unsigned char)(imageSize & 255);
    buffer[3] = (unsigned char)((imageSize >> 8) & 255);
    buffer[6] = (unsigned char)(imageSize >> 16);

    bytes = fwrite(buffer, sizeof(char), 16, out);
    if ( bytes != 16 )
            {
            fclose(in);
            fclose(out);
            return 4;
            }
    memset(buffer, 0, sizeof(buffer));
    for(i=1;i<=3;i++)
        {
        bytes = fread(buffer, sizeof(char), 128, in);
        if ( bytes != 128 )
            {
            fclose(in);
            fclose(out);
            return 3;
            }
        bytes = fwrite(buffer, sizeof(char), 256, out);
        if ( bytes != 256 )
            {
            fclose(in);
            fclose(out);
            return 4;
            }
        }
    while ( 1 )
        {
        bytes = fread(buffer, sizeof(char), sizeof(buffer), in);
        if ( !bytes )
            break;
        i = fwrite(buffer, sizeof(char), bytes, out);
        if ( bytes != i )
            {
            fclose(in);
            fclose(out);
            return 4;
            }
        if ( bytes != sizeof(buffer) )
            break;
        }
    if ( !feof(in) )
        {
        fclose(in);
        fclose(out);
        return 3;
        }
    fclose(in);
    fclose(out);
    return 0;
    }

  
/* converts a good ATR (Sio2PC, XLIT/ others) to bad
   type double density ATR (Ape/XFormer) */

int GoodAtrtoBadAtr( char *inFile , char *outFile)
    {
    int bytes,i;
    unsigned char buffer[256];
    unsigned long imageSize;
    FILE *in, *out;
    in = fopen(inFile, "rb");
    if ( !in )
        return 1;
    out = fopen(outFile, "wb");
    if ( !out )
        {
        fclose(in);
        return 2;
        }
    bytes = fread(buffer, sizeof(char), 16, in);
    if ( bytes != 16 )
            {
            fclose(in);
            fclose(out);
            return 3;
            }
    // correct the #of paragraphs in the header
    
    imageSize = 16L * (buffer[2] | buffer[3] * 256L | buffer[6] * 65536L);
    imageSize -= 384;
    imageSize /= 16;
    buffer[2] = (unsigned char)(imageSize & 255);
    buffer[3] = (unsigned char)((imageSize >> 8) & 255);
    buffer[6] = (unsigned char)(imageSize >> 16);

    bytes = fwrite(buffer, sizeof(char), 16, out);
    if ( bytes != 16 )
            {
            fclose(in);
            fclose(out);
            return 4;
            }
    memset(buffer, 0, sizeof(buffer));
    for(i=1;i<=3;i++)
        {
        bytes = fread(buffer, sizeof(char), 256, in);
        if ( bytes != 256 )
            {
            fclose(in);
            fclose(out);
            return 3;
            }
        bytes = fwrite(buffer, sizeof(char), 128, out);
        if ( bytes != 128 )
            {
            fclose(in);
            fclose(out);
            return 4;
            }
        }
    while ( 1 )
        {
        bytes = fread(buffer, sizeof(char), sizeof(buffer), in);
        if ( !bytes )
            break;
        i = fwrite(buffer, sizeof(char), bytes, out);
        if ( bytes != i )
            {
            fclose(in);
            fclose(out);
            return 4;
            }
        if ( bytes != sizeof(buffer) )
            break;
        }
    if ( !feof(in) )
        {
        fclose(in);
        fclose(out);
        return 3;
        }
    fclose(in);
    fclose(out);
    return 0;
    }



/****************************************************************
  DCM Conversion Routines
****************************************************************/

static int DecodeDcmBlock(CAtrFile * atr);
static int DecodeC1( void );
static int DecodeC3( void );
static int DecodeC4( void );
static int DecodeC6( void );
static int DecodeC7( void );
static int DecodeFA( void );

unsigned char sectorBuffer[256];
FILE *in;
unsigned short curSec;
unsigned short sectorCount, secSize;



int DCMtoATR( const char *inFile, const char *outFile)
    {
    CAtrFile atr;
    int bytes, density, stat;
    unsigned char flag;
    in = fopen( inFile, "rb");
    if ( !in )
        return 1;

    bytes = fread(&flag, 1, 1, in);  /* add header check later */

    bytes = fread(&flag, 1, 1, in);
    if ( !bytes )
        {
        fclose(in);
        return 1;
        }

   density=((flag & 0x70) >> 4);
   switch(density)
       {
       case 0:
          sectorCount=720;
          secSize=128;
          break;
       case 2:
          sectorCount=720;
          secSize=256;
          break;
       case 4:
          sectorCount=1040;
          secSize=128;
          break;
       default:
           fclose(in);
           return 3;
           break;
       }
    if ( atr.Create( (char *) outFile, sectorCount, secSize) )
        {
        fclose(in);
        return 4;
        }
    if ( fseek( in, 0, SEEK_SET) )
        {
        fclose(in);
        return 4;
        }
    fclose(in);
    in = fopen( inFile, "rb");
    if ( !in )
        return 1;

    curSec = 1;
    stat = 0;
    while ( !stat && curSec <= sectorCount )
        {
        stat = DecodeDcmBlock( &atr );
        }
    fclose(in);
    atr.Close();
    return stat;
    }

static int DecodeDcmBlock(CAtrFile * atr)
   {
   unsigned short oldSec;
   int bytes,stat = 0;
   unsigned char c;

   // get the block type
   bytes = fread(&c, 1, 1, in);
   if ( !bytes )
      return 1;

   switch(c)
      {
      case 0xFA:
      case 0xF9:
         stat = (DecodeFA() );
         return 0;
      case 0x41:
      case 0xC1:
         stat = (DecodeC1() );
         break;
      case 0x43:
      case 0xC3:
         stat = (DecodeC3() );
         break;
      case 0x44:
      case 0xC4:
         stat = (DecodeC4() );
         break;
      case 0x46:
      case 0xC6:
         stat = (DecodeC6() );
         break;
      case 0x47:
      case 0xC7:
         stat = (DecodeC7() );
         break;
      case 0x45:
         return 0; /* end of block */
      default:
         return -c;
      }
      stat = 0;
      if ( !(c&128) )
         {
         oldSec = curSec+1;
         bytes = fread(&curSec, 1, 2, in);
         if ( bytes != 2 )
            return 3;
         memset(sectorBuffer, 0, sizeof(sectorBuffer));
         while (  !stat && oldSec < curSec )
             {
             stat = atr->WriteSector(oldSec, sectorBuffer);
             oldSec++;
             }
         }
      else
          {
          stat = atr->WriteSector(curSec++, sectorBuffer);
          }

      return stat;
   }




/* 41/C1: end of sector filled with constant value, first part of
          sector verbatim data (written reversed) */

static int DecodeC1( void)
   {
   unsigned char byte[2];
   int c, bytes, offset, i;

   bytes = fread(byte, 1, 2, in);
   if ( bytes != 2 )
      return 5;
   offset = byte[0]?(int)byte[0]:256;
   c=byte[1];

   if ( offset > secSize )
      return 6;

   for (i=offset; i<(int)secSize; i++)
      sectorBuffer[i] = (unsigned char)c;

   for (offset--; offset >=0 ; offset--) 
      {
      bytes = fread(&sectorBuffer[offset], 1, 1, in);
      if ( !bytes )
         return 7;
      }
   return 0;
   }


/* RLE */
static int DecodeC3(void)
{
   unsigned char byte;
   int  offset = 0, tempOffset, bytes;

   do {
      bytes = fread(&byte, 1, 1, in);
      if ( !bytes )
         return 8;

      if (offset)
         tempOffset = byte?(int)byte:256;
      else
         tempOffset= byte; /* ???? */

      if ( tempOffset <= offset )
         return 9;

      bytes = fread(&sectorBuffer[offset], 1, tempOffset - offset, in);
      if ( bytes != tempOffset - offset )
         return 10;
      offset = tempOffset;

      if (offset == (int)secSize)
         break;
      bytes = fread(&byte, 1, 1, in);
      if ( !bytes )
         return 11;
      tempOffset=byte?(int)byte:256;

      bytes = fread(&byte, 1, 1, in);
      if ( !bytes )
         return 12;
      if ( tempOffset <= offset )
         return 13;
      for (; offset<tempOffset; offset++) 
         sectorBuffer[offset] = byte;
      } while(offset < (int)secSize);

   return 0;
}


/* 44/C4 - end of sector is verbatim data, first part of sector is same as
           last sector up to offset specified */

static int DecodeC4(void)
{
   unsigned char byte;
   int bytes, offset;
   
   bytes = fread(&byte, 1, 1, in);
   if ( !bytes )
      return 14;

   offset = byte?(int)byte:256;

   if ( offset >= secSize )
      return 15;

   bytes = fread(&sectorBuffer[offset], 1, secSize - offset, in);

   if ( bytes != (int)(secSize - offset) )
      return 16;

   return 0;
}

/* 46/C6 - sector is same as previous sector */

static int DecodeC6(void)
   {
   return 0;
   }

/* 47/C7 - Sector is not compressed, just exact data */

static int DecodeC7(void)
{
   int bytes;

   bytes = fread(sectorBuffer, 1, secSize, in);
   if ( bytes != (int)secSize )
        return 23;
   else
        return 0;
}

/* F9/FA Block/File start */

static int DecodeFA(void)
   {
   unsigned char c;
   int density, bytes;
   int last;

   bytes = fread(&c, 1, 1, in);
   if ( !bytes )
      return 1;
//   density=((c & 0x70) >> 4);
//   last=((c & 0x80) >> 7);
   density=(((unsigned short)c & 0x70) >> 4 );
   last=(((unsigned short)c & 0x80) > 127 );
   switch(density)
      {
      case 0:
         sectorCount=720;
         secSize=128;
         break;
      case 2:
         sectorCount=720;
         secSize=256;
         break;
      case 4:
         sectorCount=1040;
         secSize=128;
         break;
      default:
         return 1000+density;
      }

   bytes = fread(&curSec, 1, 2, in);
   if ( bytes != 2 )
      return 1;
   return 0;
   }


/********************************************************************
 IsBootAtr - Returns 0 if atr is a K Boot ATR
********************************************************************/

int IsBootAtr( char *atrName )
   {
   int bytes;
   FILE *inFile;
   unsigned char buffer[384+16];
   inFile = fopen( atrName, "rb");
   if ( !inFile )
       return -1;
   bytes = fread( buffer, 1, sizeof(buffer), inFile);
   if ( bytes != sizeof(buffer) )
       {
       fclose(inFile);
       return -2;
       }
   fclose(inFile);
   if (buffer[0] != 0x96 )
       return -3;
   if (buffer[1] != 0x2 )
       return -3;
   buffer[9+16] = bootData[9];
   buffer[10+16] = bootData[10];
   buffer[11+16] = bootData[11];
   buffer[12+16] = bootData[12];
   if ( !memcmp(buffer+16, bootData, 384) )
       return 0;
   else
       return 1;
   }

/********************************************************************
 CreateBootAtr - creates a minimally sized bootable ATR image from
                 an atari executable.  The executable must not need
                 DOS to run.
********************************************************************/

int CreateBootAtr( char *atrName, char *fileName)
   {
   unsigned char header[16] = {0};
   FILE * atrFile, *inFile;
   unsigned long fileSize = 0;
   unsigned char buffer[384];
   unsigned long sectorCnt;
   unsigned long paras;
   size_t padding, bytes, bytes2;
   int stat;

/* get file's size */

   inFile = fopen( fileName, "rb");
   if ( !inFile )
       return 1;

   atrFile = fopen( atrName, "wb");
   if ( !atrFile )
       {
       fclose(inFile);
       return 2;
       }

/* determine number of sectors required  */

/* create ATR header */
   memset( header, 0, sizeof(header));
   // atr signature
   header[0] = (unsigned char) 0x96;
   header[1] = (unsigned char) 0x2;
   // fill in sector size (low byte)
   header[4] = 128;

   // Write the ATR Header 
   bytes = fwrite(header, 1, sizeof(header), atrFile);
   if ( bytes != sizeof(header) )
      {
      fclose(inFile) ;
      fclose(atrFile);
      return 3;
      }

// write three boot sectors
   bytes = fwrite(bootData, 1, 384, atrFile);
   if ( bytes != 384 )
      {
      fclose(inFile) ;
      fclose(atrFile);
      return 6;
      }

/* open the input file and copy/append the file's data to output file */

   bytes = sizeof(buffer);
   while (bytes == sizeof(buffer))
      {
      bytes = fread(buffer, 1, sizeof(buffer), inFile);
      if ( !bytes )
         break;
      bytes2 = fwrite(buffer, 1, bytes, atrFile);
      fileSize += bytes2;
      if ( bytes != bytes2 )
         {
         fclose(inFile);
         fclose(atrFile);
         return 7;
         }
      }
   if ( !feof(inFile) )
      {
      fclose(inFile);
      fclose(atrFile);
      return 8;
      }
  // close input file
  fclose(inFile); 

  // calcualte # of sectors and number of "paragraphs"
  
  sectorCnt = (unsigned short) ((fileSize + 127L) / 128L + 3L);
  paras = sectorCnt * 8;

  // pad output to even sector size (data has no meaning) 
   padding = (size_t) ((sectorCnt-3) * 128 - fileSize ); 
   if ( padding )
      {
      bytes = fwrite(buffer, 1, padding, atrFile);
      if ( bytes != padding )
         {
         fclose(atrFile);
         return 9;
         }
      }  
   // update paragraphs in header
   stat = fseek(atrFile, 0, SEEK_SET);
   if ( stat )
       {
       fclose(atrFile);
       return 10;
       }
   header[2] = (unsigned char) paras & 0xFF;
   header[3] = (unsigned char) (paras>>8) & 0xFF;
   header[6] = (unsigned char) (paras>>16) && 0xFF;
   bytes = fwrite(header, 1, sizeof(header), atrFile);
   if ( bytes != sizeof(header) )
      {
      fclose(atrFile);
      return 11;
      }


   // update file size for boot loader
   bootData[9]  = (unsigned char)(fileSize & 255);
   bootData[10] = (unsigned char)((fileSize >> 8) & 255);
   bootData[11] = (unsigned char)((fileSize >> 16) & 255);
   bootData[12] = 0;
   stat = fseek(atrFile, 16, SEEK_SET);
   if ( stat )
       {
       fclose(atrFile);
       return 12;
       }
// re-write three boot sectors
   bytes = fwrite(bootData, 1, 384, atrFile);
   if ( bytes != 384 )
      {
      fclose(atrFile);
      return 13;
      }

/* close output */
   fclose(atrFile);

   return 0;
   }

/********************************************************************
ExtractExeFromBootAtr - undoes a CreateBootAtr by extracting the
                        original executable
returns 0 for success, or non-zero on error.
********************************************************************/

long ExtractExeFromBootAtr( char *atrName, char *fileName)
   {
   FILE *atrFile, *exeFile;
   unsigned char *buffer;
   unsigned char hdr[16];
   size_t bytes,bytes2,readCnt;
   unsigned long  fSize, size;


if ( IsBootAtr(atrName) != 0 )
    return 1;

/* get some memory */
   buffer = (unsigned char *)malloc(384);
   if ( !buffer)
      return 0;

/* open the atr file */
   atrFile = fopen(atrName, "rb");
   if ( !atrFile )
      {
      free(buffer);
      return 2;
      }

/* read Atr header */

   if ( fread(hdr, 1, 16, atrFile) != 16 )
      {
      free(buffer);
      fclose(atrFile);
      return 3;
      }

/* read first 3 (boot) sectors */

   if ( fread(buffer, 1, 384, atrFile) != 384 )
      {
      free(buffer);
      fclose(atrFile);
      return 4;
      }

/* Get size of file to extract */
   fSize = size = ((unsigned long)buffer[9]|(((unsigned long)buffer[10])<<8)|(((unsigned long)buffer[11])<<16));

/* Open output file */
   exeFile = fopen(fileName, "wb");
   if ( !exeFile )
      {
      fclose(atrFile);
      free(buffer);
      return 5;
      }
/* copy 'size' bytes from the Atr file to the exe file */   
   bytes = 384;
   while (bytes == 384 && fSize)
      {
      readCnt = (size_t)min(384, fSize);
      bytes = fread(buffer, 1, readCnt, atrFile);
      if ( !bytes )
         break;
      bytes2 = fwrite(buffer, 1, bytes, exeFile);
      if ( bytes != readCnt )
         {
         fclose(exeFile);
         fclose(atrFile);
         free(buffer);
         return 6;
         }
      fSize -= bytes;
      }

/* clean up and get out of here */
   fclose(exeFile);
   fclose(atrFile);
   free(buffer);

   return 0;
   }

